-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Purpose: Providing address and control signals for external memory realizing a
--          reordering of the data. Based on alternating read and writes. 
--
--
-- Description: The values of generic g_reorder_seq determine the rhythm of the 
--              sequencer. The sequencer is based on the use of a dualpage memory.
--              While data is written in page 0, page 1 is read. The state machine 
--              alternated between reads and writes: read periods and write periods.
--              The number of accesses in a read and wrte period is determined by the 
--              values of the generic.
--              The generics are interpreted as follows: 
--
--              - wr_chunksize is the number of samples that are written during a write access.
--              - wr_nof_chunks is the number of write accesses performed during a write period
--              - rd_chunksize is the number of samples that are read during a read access.
--              - rd_nof_chunks is the number of read accesses performed during a read period
--              - gapsize is the address space to be skipped in between two consecutive write or
--                read periods. Gapsize can be used to create byte-alignment in the memory.
--              - nof_blocks determines the number of write and read periods before the page 
--                swap takes place. 
--               
--              The example shows the following configuration: 
--
--              wr_chunksize    : POSITIVE := 8
--              wr_nof_chunks   : POSITIVE := 1;              
--              rd_chunksize    : POSITIVE := 4;  
--              rd_nof_chunks   : POSITIVE := 2;              
--              gapsize         : NATURAL  := 0;              
--              nof_blocks      : POSITIVE := 5;              

--                    WR             RD    
--                 --------       --------
--                 |0     |       |0,0   |   rd_chunksize  
--   wr_chunksize  |      |       --------
--                 |      |       |2,1   |
--                 --------       --------
--        gapsize  |      |       |      |
--                 --------       --------
--                 |1     |       |0,1   |
--                 |      |       --------
--                 |      |       |3,0   |
--                 --------       --------
--                 |      |       |      |
--                 --------       --------
--                 |2     |       |1,0   |  
--                 |      |       --------
--                 |      |       |3,1   |
--                 --------       --------
--                 |      |       |      |
--                 --------       --------
--                 |3     |       |1,1   |
--                 |      |       --------
--                 |      |       |4,0   |
--                 --------       --------
--                 |      |       |      |
--                 --------       -------- 
--                 |4     |       |2,0   | 
--                 |      |       -------- 
--                 |      |       |4,1   | 
--                 --------       -------- 
--                 |      |       |      | 
--                 --------       -------- 
--                  page 0         page 1
--            
--              The index in the write page indicate the write period(0-4)
--              The first index in the read page indicates the read period(0-4)
--              The second index in the read page indicates the chunknumber(0-1)
--
-- Remarks:


LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.reorder_pkg.ALL;

ENTITY reorder_sequencer IS
  GENERIC (
    g_reorder_seq   : t_reorder_seq := c_reorder_seq;
    g_data_w_ratio  : POSITIVE := 4                 -- (256/64) Ratio between datawidth at the output of the write fifo and the input of the writefifo. 
 );                                                 -- Used to determine the c_address_w. 
  PORT (
    -- Clocks and reset
    dp_rst      : IN  STD_LOGIC;  -- reset synchronous with st_clk
    dp_clk      : IN  STD_LOGIC;  -- other clock domain clock
                
    en_evt      : OUT STD_LOGIC;
    wr_not_rd   : OUT STD_LOGIC;
                
    address     : OUT STD_LOGIC_VECTOR;
    burstsize   : OUT STD_LOGIC_VECTOR;
                
    done        : IN  STD_LOGIC;
    
    sync_ok_in  : IN  STD_LOGIC;
    sync_ok_out : OUT STD_LOGIC
   );
END reorder_sequencer;


ARCHITECTURE rtl OF reorder_sequencer IS
  
  CONSTANT c_blocksize       : POSITIVE := g_reorder_seq.wr_nof_chunks * (g_reorder_seq.wr_chunksize + g_reorder_seq.gapsize);  
  CONSTANT c_page_size       : POSITIVE := c_blocksize * g_reorder_seq.nof_blocks;
  CONSTANT c_nof_wr_access   : POSITIVE := g_reorder_seq.wr_nof_chunks * g_reorder_seq.nof_blocks;
  CONSTANT c_address_w       : POSITIVE := ceil_log2(2*c_page_size);    
  CONSTANT c_address_shift_w : POSITIVE := ceil_log2(g_data_w_ratio);    
  
  TYPE   state_type is (s_idle, s_write, s_first_write, s_wait_wr, s_read, s_wait_rd);

  TYPE reg_type IS RECORD
    rd_page_offset   : NATURAL;
    wr_page_offset   : NATURAL;
    ddr3_en          : STD_LOGIC;
    wr_not_rd        : STD_LOGIC;
    rd_block_offset  : NATURAL;
    rd_chunks_offset : NATURAL;
    rd_block_cnt     : NATURAL;
    rd_chunks_cnt    : NATURAL;
    wr_block_offset  : NATURAL;
    wr_chunks_offset : NATURAL;
    wr_block_cnt     : NATURAL;
    wr_chunks_cnt    : NATURAL;
    switch_cnt       : NATURAL;   -- Counter that counts the write and read accesses to determine the switch between read and write phase. 
    page_cnt         : NATURAL;   -- Counter that counts the number of write accesses to determuine the page-swap. 
    first_write      : STD_LOGIC;
    sync_ok_out      : STD_LOGIC;
    start_addr       : STD_LOGIC_VECTOR(c_address_w - 1 DOWNTO 0);
    burstsize        : STD_LOGIC_VECTOR(c_address_w - 1 DOWNTO 0);
    state            : state_type;  -- The state machine. 
  END RECORD;

  SIGNAL r, rin      : reg_type;
   
BEGIN 
    
  ---------------------------------------------------------------
  -- CHECK IF PROVIDED GENERICS ARE ALLOWED. 
  ---------------------------------------------------------------
  ASSERT NOT((g_reorder_seq.wr_nof_chunks*g_reorder_seq.wr_chunksize) /= (g_reorder_seq.rd_nof_chunks*g_reorder_seq.rd_chunksize) AND rising_edge(dp_clk)) REPORT "Total write configuration is different from total read configuration!!!" SEVERITY FAILURE;
    
  p_comb : PROCESS(r, dp_rst, done, sync_ok_in)
    VARIABLE v : reg_type;
  BEGIN
   
    v         := r;
    v.ddr3_en := '0'; 
    
    CASE r.state IS
      WHEN s_idle => 
        IF(done = '1' AND sync_ok_in = '1') THEN     
          v.first_write := '1';
          v.sync_ok_out := sync_ok_in;
          v.state       := s_first_write;  
        END IF;

      WHEN s_first_write =>  
        v.wr_not_rd  := '1'; 
        v.ddr3_en    := '1'; 
        v.start_addr := TO_UVEC(r.wr_page_offset + r.wr_block_offset + r.wr_chunks_offset, c_address_w);
        v.burstsize  := TO_UVEC(g_reorder_seq.wr_chunksize, c_address_w);
        v.switch_cnt := r.switch_cnt + 1;
        v.state      := s_wait_wr;
        
      WHEN s_write =>  
        IF(done = '1') THEN 
          v.wr_not_rd  := '1'; 
          IF(sync_ok_in = '1') THEN   -- Only write when good sync pattern on the input.
            v.ddr3_en    := '1';
          END IF;
          v.start_addr := TO_UVEC(r.wr_page_offset + r.wr_block_offset + r.wr_chunks_offset, c_address_w);                   
          v.burstsize  := TO_UVEC(g_reorder_seq.wr_chunksize, c_address_w);
          v.switch_cnt := r.switch_cnt + 1;
          v.state      := s_wait_wr; 
        END IF;

      WHEN s_wait_wr =>      
        v.page_cnt := r.page_cnt + 1;
        IF(r.wr_block_cnt = g_reorder_seq.nof_blocks-1 AND r.wr_chunks_cnt = g_reorder_seq.wr_nof_chunks-1) THEN 
          v.wr_block_offset  := 0;
          v.wr_chunks_offset := 0;
          v.wr_block_cnt     := 0;
          v.wr_chunks_cnt    := 0;
        ELSIF(r.wr_block_cnt = g_reorder_seq.nof_blocks-1) THEN 
          v.wr_block_offset  := 0;
          v.wr_chunks_offset := r.wr_chunks_offset + g_reorder_seq.wr_chunksize;
          v.wr_block_cnt     := 0; 
          v.wr_chunks_cnt    := r.wr_chunks_cnt + 1;
        ELSE 
          v.wr_block_offset := r.wr_block_offset + c_blocksize;
          v.wr_block_cnt    := r.wr_block_cnt + 1;
        END IF;          
        
        IF(r.switch_cnt = g_reorder_seq.wr_nof_chunks) THEN
          v.switch_cnt := 0; 
          v.state      := s_read;
        ELSE
          v.state := s_write;
        END IF;  
      
      WHEN s_read => 
        IF(done = '1') THEN 
          v.wr_not_rd := '0'; 
          IF( r.first_write = '0') THEN 
            v.ddr3_en := '1'; 
          END IF; 
          v.start_addr  := TO_UVEC(r.rd_page_offset + r.rd_block_offset + r.rd_chunks_offset, c_address_w); 
          v.burstsize   := TO_UVEC(g_reorder_seq.rd_chunksize, c_address_w);
          v.switch_cnt  := r.switch_cnt + 1; 
          v.state       := s_wait_rd;
          v.sync_ok_out := sync_ok_in;
        END IF;
  
      WHEN s_wait_rd => 
        IF(r.switch_cnt = g_reorder_seq.rd_nof_chunks) THEN 
          v.switch_cnt := 0; 
          v.state      := s_write;
        ELSE
          v.state := s_read;
        END IF;  

        IF(r.rd_block_cnt = g_reorder_seq.nof_blocks-1 AND r.rd_chunks_cnt = g_reorder_seq.rd_nof_chunks-1) THEN 
          v.rd_block_offset  := 0;
          v.rd_chunks_offset := 0;
          v.rd_block_cnt     := 0;
          v.rd_chunks_cnt    := 0;    
          IF(r.page_cnt = c_nof_wr_access) THEN 
            v.rd_page_offset := r.wr_page_offset;
            v.wr_page_offset := r.rd_page_offset;
            v.page_cnt       := 0;  
            v.first_write    := '0';
            IF(sync_ok_in = '0') THEN
              v.state := s_idle;
            END IF;
          END IF; 
        ELSIF(r.rd_block_cnt = g_reorder_seq.nof_blocks-1) THEN 
          v.rd_block_offset  := 0;
          v.rd_chunks_offset := r.rd_chunks_offset + g_reorder_seq.rd_chunksize;
          v.rd_block_cnt     := 0; 
          v.rd_chunks_cnt    := r.rd_chunks_cnt + 1;         
        ELSE 
          v.rd_block_offset := r.rd_block_offset + c_blocksize;
          v.rd_block_cnt    := r.rd_block_cnt + 1;
        END IF;
        
      WHEN OTHERS =>
        v.state := s_idle;

    END CASE;
    
    IF(dp_rst = '1') THEN 
      v.rd_page_offset   := c_page_size;
      v.wr_page_offset   := 0;
      v.page_cnt         := 0;
      v.switch_cnt       := 0; 
      v.ddr3_en          := '0';
      v.wr_not_rd        := '0';
      v.wr_block_offset  := 0;
      v.wr_chunks_offset := 0;
      v.wr_block_cnt     := 0;
      v.wr_chunks_cnt    := 0;
      v.rd_block_offset  := 0;
      v.rd_chunks_offset := 0;
      v.rd_block_cnt     := 0;
      v.rd_chunks_cnt    := 0;
      v.sync_ok_out      := '0';
      v.start_addr       := (OTHERS => '0');
      v.burstsize        := (OTHERS => '0');
      v.first_write      := '1';
      v.state            := s_idle;
    END IF;

    rin <= v;  
         
  END PROCESS;
  
  p_regs : PROCESS(dp_clk)
  BEGIN 
    IF RISING_EDGE(dp_clk) THEN 
      r <= rin; 
    END IF; 
  END PROCESS;

  en_evt      <= r.ddr3_en;
  wr_not_rd   <= r.wr_not_rd;
  sync_ok_out <= r.sync_ok_out;
  address     <= RESIZE_UVEC(r.start_addr(c_address_w-1 DOWNTO c_address_shift_w), address'LENGTH);
  burstsize   <= RESIZE_UVEC(r.burstsize(c_address_w-1 DOWNTO c_address_shift_w),  burstsize'LENGTH);

END rtl;

